import React, { PureComponent } from 'react';
import PropTypes from 'prop-types';
import isEqual from 'lodash.isequal';
import pick from 'lodash.pick';
import isFunction from 'lodash.isfunction';
import { Portal } from 'react-overlays';
import {
  stringify, generateItemsFromOptions,
  isSearchable, isMinLengthMatched, getResults
} from './utils';
import Menu from './Menu';
import {
  DEFAULT_LABELKEY, SELECT_HEIGHT, SELECT_WIDTH, MAXIMUM_DROPDOWN_HEIGHT, MINIMUM_DROPDOWN_WIDTH,
  UP, DOWN, ESC, RETURN, TAB,
  OPTIONS_PROPS, DROPDOWN_ITEM_HEIGHT,
  ACTUAL_VALUE
} from './constants';
import ArrowDown from './icons/arrow-down';
import TextInput from './TextInput';

function offset(el) {
  const rect = el.getBoundingClientRect().toJSON();
  const scrollLeft = window.pageXOffset || document.documentElement.scrollLeft
  const scrollTop = window.pageYOffset || document.documentElement.scrollTop;
  return {
    ...rect,
    top: rect.top + scrollTop,
    left: rect.left + scrollLeft
  };
}

const getValue = (state, props) => {
  let { selected = [], value = '' } = state;
  const { labelKey } = props;
  selected = [...selected].shift();
  if (selected && typeof selected[labelKey] !== 'undefined') {
    value = selected[labelKey];
  }
  return value;
}

const getActiveIndex = (options = [], selected = false, props) => {
  const { labelKey, defaultFirstItemSelected, async } = props;
  if (!selected) {
    return defaultFirstItemSelected && options.length ? 0 : -1;
  }
  if (async) {
    return defaultFirstItemSelected ? 0 : -1;
  }
  const index = options.findIndex(option => option[labelKey] === selected[labelKey]);
  return index;
};

const initialStateFromProps = props => {
  const options = [...generateItemsFromOptions(props.options, props.labelKey)];
  const selected = [...generateItemsFromOptions(props.selected, props.labelKey)];
  return {
    options,
    results: [...options],
    selected,
    value: getValue({ selected }, props),
    activeIndex: getActiveIndex(options, selected.length && selected[0], pick(props, ['labelKey', 'defaultFirstItemSelected', 'async'])),
  }
};

class Select extends PureComponent {
  static propTypes = {
    options: PropTypes.array.isRequired,
    selected: PropTypes.array.isRequired,
    labelKey: PropTypes.oneOfType([PropTypes.string, PropTypes.func]),
    searchable: PropTypes.bool,
    onChange: PropTypes.func.isRequired,
    placeHolder: PropTypes.string,
    closeOnOutsideClick: PropTypes.bool,
    multiple: PropTypes.bool,
    maxDropdownHeight: PropTypes.oneOfType([PropTypes.string, PropTypes.number]),
    minDropdownWidth: PropTypes.oneOfType([PropTypes.string, PropTypes.number]),
    width: PropTypes.oneOfType([PropTypes.string, PropTypes.number]),
    height: PropTypes.oneOfType([PropTypes.string, PropTypes.number]),
    minLength: PropTypes.number,
    filterBy: PropTypes.oneOfType([PropTypes.array, PropTypes.func]),
    defaultFirstItemSelected: PropTypes.bool,
    onInputChange: PropTypes.func,
    loading: PropTypes.bool,
    inputProps: PropTypes.object,
    onFocus: PropTypes.func,
    onBlur: PropTypes.func,
    clearable: PropTypes.bool,
    icons: PropTypes.object,
    dropup: PropTypes.bool,
    flip: PropTypes.bool,
    iconColor: PropTypes.string,
    primaryColor: PropTypes.string,
    defaultInputValue: PropTypes.string,
  };

  static defaultProps = {
    options: [],
    selected: [],
    labelKey: DEFAULT_LABELKEY,
    searchable: false,
    onChange: () => { },
    placeHolder: 'Select...',
    maxDropdownHeight: MAXIMUM_DROPDOWN_HEIGHT,
    minDropdownWidth: MINIMUM_DROPDOWN_WIDTH,
    width: SELECT_WIDTH,
    height: SELECT_HEIGHT,
    closeOnOutsideClick: true,
    multiple: false,
    minLength: 0,
    filterBy: [],
    defaultFirstItemSelected: false,
    onInputChange: () => { },
    loading: false,
    inputProps: {},
    onFocus: () => { },
    onBlur: () => { },
    clearable: false,
    icons: {},
    dropup: false,
    flip: false,
    iconColor: '#888',
    primaryColor: '#0057e7',
    defaultInputValue: ''
  };

  state = {
    ...initialStateFromProps(this.props),
    open: false,
    overflow: 'hidden',
    hasFocus: false,
    menuPosition: {
      display: 'none'
    },
    select: null,
    autoScroll: false,
    isDirty: false,
  };

  _inputRef = input => {
    this._input = input;
    this.setState({
      menuPosition: this.getMenuPosition()
    });
  };
  clientWidth = window.innerWidth;
  clientHeight = window.innerHeight;

  componentDidMount() {
    this.addResizeListener();
  }

  UNSAFE_componentWillReceiveProps(nextProps) {
    const { labelKey, multiple, selected, options, defaultFirstItemSelected } = nextProps;

    const nextOptions = generateItemsFromOptions([...options], labelKey);

    if (stringify(nextOptions) !== stringify(this.state.options)) {
      const results = getResults(nextOptions, nextProps, this.state);

      this.setState({
        menuPosition: this.getMenuPosition(this.calculateHeight(results.length))
      }, () => {
        this.setState({
          options: [...nextOptions],
          results: [...results],
          activeIndex: defaultFirstItemSelected ? 0 : -1
        });
      });
    }

    if (multiple !== this.props.multiple) {
      this.setState({ value: '' });
    }

    // If new selections are passed via props, treat as a controlled input.
    if (selected && !isEqual(selected, this.state.selected)) {
      this.setState({
        selected: [...generateItemsFromOptions(selected, labelKey)]
      });

      if (multiple) {
        return;
      }
      this.setState({
        value: selected.length ? selected[0][labelKey] : '',
        activeIndex: getActiveIndex(nextOptions, selected, nextProps)
      });
    }

    // Truncate selections when in single-select mode.
    let newSelected = selected || this.state.selected;
    if (!multiple && newSelected.length > 1) {
      newSelected = newSelected.slice(0, 1);
      this.setState({
        selected: [...generateItemsFromOptions(newSelected, labelKey)],
        value: newSelected[0][labelKey],
      });
    }
  }

  componentDidUpdate(prevProps, prevState) {
    if(prevState.value !== this.state.value) {
      this.setState({
        results: getResults(this.state.results, this.props, this.state)
      });
    }
  }

  componentWillUnmount() {
    this.removeResizeListener();
  }

  onResize = () => {
    if (this.isMenuOpen()) {
      this.checkMenuPosition();
    }
  }

  addResizeListener = () => {
    window.addEventListener('resize', this.onResize);
  }

  removeResizeListener = () => {
    window.removeEventListener('resize', this.onResize);
  }

  toogleDropDown = () => {
    if (this.isMenuOpen()) {
      return this.closeMenu();
    }
    this.openMenu();
  };

  setMenuRef = menu => {
    this._menu = menu;
  };

  setupEventListeners = () => {
    document.addEventListener('click', this.handleOutsideClick, false);
  };

  revokeEventListeners = () => {
    document.removeEventListener('click', this.handleOutsideClick, false);
  };

  handleOutsideClick = event => {
    if (this.isMenuOpen() && (this._menu && !this._menu.contains(event.target)) && (this._input !== event.target) && (this._icon && !this._icon.contains(event.target)) && this.props.closeOnOutsideClick) {
      this.closeMenu();
    }
  };

  checkMenuPosition = (debug) => {
    const menuPosition = this.getMenuPosition(debug);
    if (stringify(menuPosition) !== stringify(this.state.menuPosition)) {
      this.setState({
        menuPosition
      });
    }
  }

  calculateHeight = (length) => Math.min(this.props.maxDropdownHeight, (length ? (length * DROPDOWN_ITEM_HEIGHT) + 15 : DROPDOWN_ITEM_HEIGHT));

  getMenuPosition = (height) => {
    if (!this._input) return {};
    const menuHeight = height || this.calculateHeight(this.state.results.length);
    const body = document.documentElement || document.body;
    const selectPosition = offset(this._input);
    let bottom = 'auto';
    let top = selectPosition.top + selectPosition.height + 2;
    let dropup = false;
    if (this.props.dropup || (this.props.flip && ((selectPosition.bottom + menuHeight) > (body.scrollHeight - body.scrollTop)))) {
      top = 'auto';
      bottom = (body.scrollHeight - body.scrollTop - selectPosition.top + 2);
      dropup = true;
    }
    let ret = {
      top,
      bottom,
      left: selectPosition.left,
      width: selectPosition.width,
      dropup
    };
    return ret;
  };

  autoScroll = () => this.state.autoScroll;

  enableAutoScroll = () => {
    if (!this.autoScroll()) {
      this.setState({
        autoScroll: true
      });
    }
  };

  disableAutoScroll = () => {
    if (this.autoScroll()) {
      this.setState({
        autoScroll: false
      });
    }
  };

  hasSelectedValue = () => {
    let { selected, labelKey } = this.props;
    selected = [...selected].shift();
    return selected && selected[labelKey];
  };

  _handleChange = selected => {
    this.props.onChange && this.props.onChange(selected);
  }

  handleInputChange = e => {
    e.persist();
    const value = e.target.value;
    const { onInputChange, multiple } = this.props;
    const clearSelected = this.state.selected.length && !multiple;
    this.setState((state) => {
      return {
        isDirty: true,
        selected: clearSelected ? [] : state.selected,
        value,
      }
    }, () => {
      isMinLengthMatched(value, this.props) && this.openMenu();
      onInputChange(value, e);
      clearSelected && this._handleChange([]);
    });
  }

  handleKeyDown = (e, results) => {
    let { activeIndex } = this.state;
    switch (e.keyCode) {
      case UP:
      case DOWN:
        if (!this.isMenuOpen()) {
          this.openMenu();
          return;
        }
        // Prevents input cursor from going to the beginning when pressing up.
        e.preventDefault();
        // Increment or decrement index based on user keystroke.
        activeIndex += e.keyCode === UP ? -1 : 1;
        // If we've reached the end, go back to the beginning or vice-versa.
        if (activeIndex === results.length) {
          activeIndex = -1;
        } else if (activeIndex === -2) {
          activeIndex = results.length - 1;
        }
        this.enableAutoScroll();
        this.handleActiveIndexChange(activeIndex);
        break;
      case ESC:
        this.isMenuOpen() && this.closeMenu();
        break;
      case RETURN:
        if (!this.isMenuOpen()) return;
        // Prevent form submission while menu is open.
        e.preventDefault();
        if (activeIndex !== -1 && results[activeIndex]) {
          this.handleMenuSelect(results[activeIndex]);
        }
        break;
      case TAB:
        if (!this.isMenuOpen()) {
          break;
        }
        this.closeMenu();
        break;
      default:
        break;
    }
  };

  handleActiveIndexChange = activeIndex => {
    const newState = { activeIndex };
    this.setState(newState);
  };

  handleMenuSelect = (option) => {
    const { onChange } = this.props;
    const {
      [ACTUAL_VALUE]: selected,
    } = option;
    onChange([selected]);
    this.closeMenu();
  };

  hasFocus = () => this.state.hasFocus;

  handleFocus = e => {
    e.persist();
    if (this.hasFocus()) return;
    this.setState({
      hasFocus: true
    }, () => {
      const value = getValue(this.state, this.props);
      isSearchable(value, this.props) && this.openMenu();
      isFunction(this.props.onFocus) && this.props.onFocus(e);
    });
  };

  handleBlur = e => {
    e.persist();
    if (!this.hasFocus()) return;
    this.setState({
      hasFocus: false
    }, () => {
      isFunction(this.props.onBlur) && this.props.onBlur(e);
    });
  };

  focus = () => {
    this._input && this._input.focus();
  };

  blur = () => {
    this._input && this._input.blur();
  };

  isMenuOpen = () => this.state.open;

  forceToggle = () => {
    if (!this.isMenuOpen()) {
      this.focus();
      this.openMenu();
    } else {
      this.blur();
      this.closeMenu();
    }
  }

  openMenu = () => {
    this.checkMenuPosition();
    this.enableAutoScroll();
    const {
      options,
      selected
    } = this.state;
    const activeIndex = getActiveIndex(options, selected.length && selected[0], pick(this.props, ['labelKey', 'multiple', 'defaultFirstItemSelected', 'async']));
    this.setState({
      open: true,
      activeIndex
    }, () => {
      setTimeout(() => {
        this.setupEventListeners();
      }, 0);
    });
  };

  closeMenu = () => {
    this.disableAutoScroll();
    this.setState({
      open: false,
      forceOpen: false,
      activeIndex: -1,
      clearValue: true,
    }, () => {
      setTimeout(() => {
        this.revokeEventListeners();
      }, 0);
    });
  };

  clear = () => {
    this.props.onChange([]);
    this.setState({
      value: ''
    });
  }

  render() {
    const {
      height, maxDropdownHeight, searchable, // clearable,
      width, placeHolder: placeholder, loading, icons
    } = this.props;
    const { overflow, hasFocus, menuPosition, results, value } = this.state;
    const menuProps = pick(this.state, OPTIONS_PROPS);
    // const hasValue = this.state.selected && this.state.selected.length;
    menuProps.loading = loading;
    const iconProps = {
      ref: icon => this._icon = icon,
    };
    const inputProps = {
      ...this.props.inputProps,
      isMenuOpen: this.isMenuOpen(),
      type: 'text',
      placeholder,
      value,
      readOnly: !searchable,
      inputRef: this._inputRef,
      onFocus: this.handleFocus,
      onBlur: this.handleBlur,
      onKeyDown: e => this.handleKeyDown(e, results)
    };
    if (!searchable) {
      iconProps.onClick = () => this.forceToggle();
      inputProps.onClick = () => this.toogleDropDown();
    } else {
      iconProps.onClick = () => this.forceToggle();
      inputProps.onChange = this.handleInputChange;
    }
    if (!icons.right) {
      icons.right = {
        props: {
          ...iconProps
        },
        additionalClasses: 'is-clickable'
      }
      if (icons.dropdown) {
        icons.right.component = icons.dropdown;
      } else {
        icons.right.component = <ArrowDown />;
      }
    }
    return (
      <div className={`pk-select${hasFocus ? ' pk-select-focused' : ''}`} style={{ height, overflow, width }}>
        <TextInput
          containerClass='pk-select--container'
          icon={icons}
          loading={loading}
          {...inputProps}
        />
          <Portal container={document.body}>
            <Menu
              options={results}
              labelKey={this.props.labelKey}
              onEntering={this.whileDropDownShow}
              onExiting={this.whileDropDownHide}
              onEnter={this.afterDropDownShow}
              onExit={this.afterDropDownHide}
              maxHeight={maxDropdownHeight}
              style={menuPosition}
              menuRef={this.setMenuRef}
              onMenuSelect={this.handleMenuSelect}
              disableAutoScroll={this.disableAutoScroll}
              {...menuProps}
            />
          </Portal>
        </div>
      // </div>
    );
  }
}

export default Select;